/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.monitor.ri.client;

import java.util.HashMap;
import java.util.Map;
import javax.swing.event.EventListenerList;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.ImplementationException;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.Properties;
import org.jdtaus.core.container.Property;
import org.jdtaus.core.container.PropertyException;
import org.jdtaus.core.monitor.Task;
import org.jdtaus.core.monitor.TaskEvent;
import org.jdtaus.core.monitor.TaskListener;
import org.jdtaus.core.monitor.spi.TaskMonitor;

/**
 * jDTAUS Core SPI Client {@code TaskMonitor} implementation.
 * <p>This client implementation will start a new monitoring thread when
 * requested with the {@code monitor()} method polling the state of the task
 * every {@code pollIntervalMillis} milliseconds to fire corresponding
 * {@code TaskEvent}s.</p>
 * <p><b>Note:</b><br/>
 * {@code TaskEvent}s produced by this implementation will be fired from the
 * monitor thread, not from the thread executing the task.</p>
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: ClientTaskMonitor.java 3463 2007-05-16 00:35:33Z schulte2005 $
 *
 * @see org.jdtaus.core.container.Container
 * @see #initialize()
 */
public final class ClientTaskMonitor
    implements TaskMonitor, ContainerInitializer
{

    //--Implementation----------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /** Meta-data describing the implementation. */
    private static final Implementation META =
        ModelFactory.getModel().getModules().
        getImplementation(ClientTaskMonitor.class.getName());

    //----------------------------------------------------------Implementation--
    //--Constructors------------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /**
     * Protected <code>ClientTaskMonitor</code> implementation constructor.
     *
     * @param meta Implementation meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    protected ClientTaskMonitor(final Implementation meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }
    /**
     * Protected <code>ClientTaskMonitor</code> dependency constructor.
     *
     * @param meta dependency meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    protected ClientTaskMonitor(final Dependency meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }

    /**
     * Initializes the properties of the instance.
     *
     * @param meta the property values to initialize the instance with.
     *
     * @throws NullPointerException if {@code meta} is {@code null}.
     */
    protected void initializeProperties(final Properties meta)
    {
        Property p;

        if(meta == null)
        {
            throw new NullPointerException("meta");
        }

        p = meta.getProperty("pollIntervalMillis");
        this._pollIntervalMillis = ((java.lang.Long) p.getValue()).longValue();

    }

    //------------------------------------------------------------Constructors--
    //--ContainerInitializer----------------------------------------------------

    /**
     * Initializes the instance.
     *
     * @throws PropertyException for invalid property values.
     *
     * @see #assertValidProperties()
     */
    public void initialize()
    {
        this.assertValidProperties();
    }

    //----------------------------------------------------ContainerInitializer--
    //--Properties--------------------------------------------------------------

    // This section is managed by jdtaus-container-mojo.

    /**
     * Property {@code pollIntervalMillis}.
     * @serial
     */
    private long _pollIntervalMillis;

    /**
     * Gets the value of property <code>pollIntervalMillis</code>.
     *
     * @return the value of property <code>pollIntervalMillis</code>.
     */
    protected long getPollIntervalMillis()
    {
        return this._pollIntervalMillis;
    }


    //--------------------------------------------------------------Properties--
    //--TaskEventSource---------------------------------------------------------

    /** List of {@code TaskListener}s. */
    private final EventListenerList taskListeners = new EventListenerList();

    public void addTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        this.taskListeners.add(TaskListener.class, listener);
    }

    public void removeTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        this.taskListeners.remove(TaskListener.class, listener);
    }

    public TaskListener[] getTaskListeners()
    {

        return (TaskListener[]) this.taskListeners.
            getListeners(TaskListener.class);

    }

    //---------------------------------------------------------TaskEventSource--
    //--TaskMonitor-------------------------------------------------------------

    /** Maps {@code Task}s to theire monitoring threads. */
    private final Map monitorMap = new HashMap(100);

    public void monitor(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        final MonitorThread t;
        synchronized(this.monitorMap)
        {
            this.finish(task);
            t = new MonitorThread(task, this.getPollIntervalMillis());
            this.monitorMap.put(task, t);
            t.start();
        }
    }

    public void finish(Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        final MonitorThread t;
        synchronized(this.monitorMap)
        {
            if((t = (MonitorThread) this.monitorMap.remove(task)) != null)
            {
                try
                {
                    t.setRunning(false);
                    t.interrupt();
                    t.join();
                }
                catch(InterruptedException e)
                {
                    throw new ImplementationException(
                        ClientTaskMonitor.META, e);

                }
            }
        }
    }

    //-------------------------------------------------------------TaskMonitor--
    //--ClientTaskMonitor-------------------------------------------------------

    /* The next thread's number. */
    private static int threadNumber;
    private static synchronized int nextThreadNumber()
    {
        return threadNumber++;
    };

    /** Thread monitoring a {@code Task}. */
    public class MonitorThread extends Thread
    {

        private boolean running;
        private final long pollIntervalMillis;
        private final Task task;

        // Task state.
        private boolean indeterminate;
        private boolean cancelable;
        private boolean cancelled;
        private int minimum;
        private int maximum;
        private int progress;

        /**
         * Creates a new {@code MonitorThread} instance.
         *
         * @param task the task to monitor.
         * @param pollIntervalMillis amount of milliseconds to wait before
         * checking the state of {@code task} for changes.
         *
         * @throws NullPointerException if {@code task} is {@code null}.
         * @throws IllegalArgumentException if {@code pollIntervalMillis} is
         * negative or equal to zero.
         */
        protected MonitorThread(final Task task,
            final long pollIntervalMillis)
        {
            super("ClientTaskMonitor-" + nextThreadNumber());

            if(task == null)
            {
                throw new NullPointerException("task");
            }
            if(pollIntervalMillis <= 0L)
            {
                throw new IllegalArgumentException(
                    Long.toString(pollIntervalMillis));

            }

            this.task = task;
            this.running = true;
            this.pollIntervalMillis = pollIntervalMillis;
            this.indeterminate = task.isIndeterminate();
            this.cancelable = task.isCancelable();

            if(!indeterminate)
            {
                this.minimum = task.getMinimum();
                this.maximum = task.getMaximum();
                this.progress = task.getProgress();
            }
            else
            {
                this.minimum = Integer.MIN_VALUE;
                this.maximum = Integer.MIN_VALUE;
                this.progress = Integer.MIN_VALUE;
            }

            if(cancelable)
            {
                this.cancelled = task.isCancelled();
            }
        }

        /** {@inheritDoc} */
        public void run()
        {
            fireTaskEvent(new TaskEvent(this.task, TaskEvent.STARTED));
            while(this.isRunning())
            {
                try
                {
                    Thread.sleep(this.pollIntervalMillis);
                    if(this.changedState())
                    {
                        fireTaskEvent(new TaskEvent(this.task,
                            TaskEvent.CHANGED_STATE));

                    }
                }
                catch(InterruptedException e)
                {
                    if(this.changedState())
                    {
                        fireTaskEvent(new TaskEvent(this.task,
                            TaskEvent.CHANGED_STATE));

                    }
                    this.setRunning(false);
                }
            }
            fireTaskEvent(new TaskEvent(this.task, TaskEvent.ENDED));
        }

        /**
         * Getter for property {@code running}.
         *
         * @return {@code true} if the thread is currently running;
         * {@code false} if not.
         */
        public boolean isRunning()
        {
            return this.running;
        }

        /**
         * Setter for property {@code running}.
         * <p/><b>Note:</b><br/>
         * To stop a thread the instance needs to be interrupted after setting
         * this property to {@code false}.
         *
         * @param running {@code true} to mark the thread as currently running;
         * {@code false} to stop the thread.
         */
        public void setRunning(final boolean running)
        {
            this.running = running;
        }

        /**
         * Checks the state of the task for changes.
         *
         * @return {@code true} if the state of the task held by the instance
         * changed since the last time this method got called; {@code false} if
         * the state did not change.
         */
        protected boolean changedState()
        {
            boolean changedState = false;
            if(this.indeterminate)
            {
                this.indeterminate = this.task.isIndeterminate();
                if(!this.indeterminate)
                {
                    this.minimum = this.task.getMinimum();
                    this.maximum = this.task.getMaximum();
                    this.progress = this.task.getProgress();
                    changedState = true;
                }
            }
            else
            {
                this.indeterminate = this.task.isIndeterminate();
                if(this.indeterminate)
                {
                    changedState = true;
                }
                else
                {
                    if(this.minimum != this.task.getMinimum())
                    {
                        this.minimum = this.task.getMinimum();
                        changedState = true;
                    }
                    if(this.maximum != this.task.getMaximum())
                    {
                        this.maximum = this.task.getMaximum();
                        changedState = true;
                    }
                    if(this.progress != this.task.getProgress())
                    {
                        this.progress = this.task.getProgress();
                        changedState = true;
                    }
                }
            }

            if(this.cancelable)
            {
                this.cancelable = this.task.isCancelable();
                if(!this.cancelable)
                {
                    changedState = true;
                }
                else
                {
                    if(this.cancelled != this.task.isCancelled())
                    {
                        this.cancelled = this.task.isCancelled();
                        changedState = true;
                    }
                }
            }
            else
            {
                this.cancelable = this.task.isCancelable();
                if(!this.cancelable)
                {
                    this.cancelled = false;
                    changedState = true;
                }
            }

            return changedState;
        }
    }

    /** Creates a new {@code ClientTaskMonitor} instance. */
    public ClientTaskMonitor()
    {
        this(ClientTaskMonitor.META);
        this.initialize();
    }

    /**
     * Checks configured properties.
     *
     * @throws PropertyException for illegal property values.
     */
    protected void assertValidProperties()
    {
        if(this.getPollIntervalMillis() < 100L)
        {
            throw new PropertyException("pollIntervalMillies",
                Long.toString(this.getPollIntervalMillis()));

        }
    }

    /**
     * Notifies all registered {@code TaskListener}s about {@code TaskEvent}s.
     *
     * @param e The event to be provided to the listeners.
     */
    protected void fireTaskEvent(final TaskEvent e)
    {
        if(e == null)
        {
            throw new NullPointerException("e");
        }

        synchronized(this.taskListeners)
        {
            final Object[] listeners = this.taskListeners.getListenerList();
            for (int i = listeners.length - 2; i >= 0; i -= 2)
            {
                if (listeners[i] == TaskListener.class)
                {
                    ((TaskListener) listeners[i + 1]).onTaskEvent(e);
                }
            }
        }
    }

    //-------------------------------------------------------ClientTaskMonitor--

}
